﻿using System;
using System.Text;
using System.Collections.Generic;
using System.Linq;
using System.Transactions;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace DBTesting
{
    /// <summary>
    /// Tests for Orders table
    /// </summary>
    [TestClass]
    public class OrderTests
    {
        public OrderTests()
        {
        }
        
        /// <summary>
        /// Gets or sets database context used by LINQ.
        /// </summary>
        private DataClassesDataContext DBContext;
        
        /// <summary>
        /// SQL transaction scope. When disposed rolls back the transaction.
        /// Used here to roll back any database changes introduced by tests.
        /// </summary>
        private TransactionScope Transaction;

        // Use TestInitialize to run code before running each test 
        [TestInitialize()]
        public void MyTestInitialize() 
        {
            Transaction = new TransactionScope();
            DBContext = new DataClassesDataContext();
        }
        
        // Use TestCleanup to run code after each test has run
        [TestCleanup()]
        public void MyTestCleanup() 
        {
            Transaction.Dispose();
        }

        //***************************************************************
        // Table Design – Match your objects?

        [TestMethod]
        [Description("Ensure the Orders table has all expected columns")]
        public void TestOrdersColumns()
        {
            const string sql = "SELECT ID, OrderDate, Customer FROM Orders";
            DBContext.ExecuteCommand(sql);
        }

        [TestMethod]
        [Description("Ensure the Orders table has all expected columns")]
        public void TestOrderLineColumns()
        {
            const string sql = "SELECT OrderID, ProductName, Count, Price FROM OrderLine";
            DBContext.ExecuteCommand(sql);
        }

        [TestMethod]
        [Description("Ensure that invalid SQL code throws exceptions")]
        [ExpectedException(typeof(System.Data.SqlClient.SqlException))]
        public void TestInvalidQueryThrows()
        {
            const string sql = "SELECT Nonexisting FROM Orders";
            DBContext.ExecuteCommand(sql);
            Assert.Fail("Expected exception");
        }

        //***************************************************************
        // Table insert

        [TestMethod]
        [Description("Ensure that orders can be created.")]
        public void TestInsertOrder()
        {
            DateTime orderDate = DateTime.Now;
            string customer = "Sven 1234567890123456789012345678901234567890 Rosvall";

            const string insertSql = "INSERT INTO Orders (OrderDate, Customer) " +
                "VALUES ({0}, {1})";
            DBContext.ExecuteCommand(insertSql, orderDate, customer);

            const string querySql = "SELECT ID, OrderDate, Customer FROM Orders "+
                "WHERE OrderDate = {0}";
            var result = DBContext.ExecuteQuery<Order>(querySql, orderDate).ToArray();

            Assert.AreEqual(1, result.Count());
            Assert.AreEqual(customer, result.First().Customer);
            TimeSpan diff = orderDate.Subtract(result.First().OrderDate);
            Assert.IsTrue(Math.Abs(diff.TotalSeconds) <= 1);
        }

        //******************************************************************************
        // Auto-increment

        [TestMethod]
        [Description("Ensure that ID is bumped up on insert")]
        public void TestIdentityIncrement()
        {
            // Insert two entries
            int id1 = InsertOrder("Peter Gabriel");
            int id2 = InsertOrder("Phil Collins");


            // Verify the new entry ID is one higher than the previous.
            Assert.AreEqual(id1+1, id2);
        }

        //******************************************************************************
        // Constraints – Foreign Key

        [TestMethod]
        [Description("Ensure no OrderLine rows can be inserted without a parent Order (Foreign Key constraint)")]
        [ExpectedException(typeof(System.Data.SqlClient.SqlException))]
        public void TestInsertOrderLineFailsWithoutOrder()
        {
            const string sql = "INSERT INTO OrderLine (OrderID, ProductName, Count, Price) "+
                "VALUES (-1, 'Sand Grain', 1000000, 10)";
            DBContext.ExecuteCommand(sql);
        }

        [TestMethod]
        [Description("Ensure no OrderLine rows can be inserted without a parent Order")]
        public void TestInsertOrderLineSucceedsWithOrder()
        {
            int orderID = InsertOrder("Sandman");

            const string sql = "INSERT INTO OrderLine (OrderID, ProductName, Count, Price) " +
                "VALUES ({0}, 'Sand Grain', 1000000, 10)";
            DBContext.ExecuteCommand(sql, orderID);
        }

        //******************************************************************************
        // Constraints – Not Null

        [TestMethod]
        [Description("Ensure Customer must be provided for inserts to Orders")]
        [ExpectedException(typeof(System.Data.SqlClient.SqlException))]
        public void TestInsertOrderWithoutCustomer()
        {
            DateTime orderDate = DateTime.Now;
            const string insertSql = "INSERT INTO Orders (OrderDate) " +
                "VALUES ({0})";
            DBContext.ExecuteCommand(insertSql, orderDate);
        }

        //******************************************************************************
        // Constraints – Default Values

        [TestMethod]
        [Description("Ensure omitted OrderDate is filled in with current date.")]
        public void TestInsertOrderWithoutOrderDate()
        {
            const string customer = "Sid Vicious";
            DateTime orderDate = DateTime.Now;
            const string insertSql = "INSERT INTO Orders (Customer) " +
                "VALUES ({0})";
            DBContext.ExecuteCommand(insertSql, customer);

            const string querySql = "SELECT ID, OrderDate, Customer FROM Orders " +
                "WHERE Customer = {0}";
            var result = DBContext.ExecuteQuery<Order>(querySql, customer).ToArray();
            Assert.AreEqual(1, result.Count());
            Assert.AreEqual(customer, result.First().Customer);
            TimeSpan diff = orderDate.Subtract(result.First().OrderDate);
            Assert.IsTrue(Math.Abs(diff.TotalSeconds) <= 10);
        }

        //******************************************************************************
        // Constraints – Value ranges

        [TestMethod]
        [Description("Ensure that inserted order lines have a count >0")]
        [ExpectedException(typeof(System.Data.SqlClient.SqlException))]
        public void TestInsertOrderLineWithZeroCount()
        {
            // Create an order for the order line
            const string customer = "Frank Zappa";
            int orderID = InsertOrder(customer);

            const string insertSql = "INSERT INTO OrderLine (OrderID, ProductName, Count, Price) " +
                "VALUES({0}, 'Guitar', 0, 0)";
            DBContext.ExecuteCommand(insertSql, orderID);

            Assert.Fail("Expect insert with zero count to fail.");
        }

        //******************************************************************************
        // Stored Procedures

        [TestMethod]
        [Description("Ensure stored procedure CreateOrderLine creates rows in Order and OrderLine")]
        public void TestCreateOrderLineForNonExistingOrder()
        {

            const string customer = "Eric Clapton";
            const string product = "Plectrum";

            const string searchOrderSql = "SELECT * FROM Orders WHERE Customer = {0}";
            const string searchOrderLineSql = "SELECT * FROM OrderLine WHERE OrderID = {0}";

            // Ensure that we start without orders
            Assert.AreEqual(0, DBContext.ExecuteQuery<Order>(searchOrderSql, customer).Count());

            // Create the order
            DBContext.CreateOrderLine(customer, product, 1, 17);

            // Ensure that we have one order now
            var orders = DBContext.ExecuteQuery<Order>(searchOrderSql, customer).ToArray();
            Assert.AreEqual(1, orders.Count());

            // Ensure that we have one order line
            int orderID = orders.First().ID;
            var orderLines = DBContext.ExecuteQuery<OrderLine>(searchOrderLineSql, orderID).ToArray();
            Assert.AreEqual(1, orderLines.Count());
            Assert.AreEqual(product, orderLines.First().ProductName);
            Assert.AreEqual(17, orderLines.First().Price);
        }
        
        //******************************************************************************
        // Constraints – Unique
        // Triggers – Cascading delete
        // Views
        // Access
        // Partitioned Tables

        //******************************************************************************
        // Helper functions

        private int InsertOrder(string customer, DateTime orderDate)
        {
            const string insertSql = "INSERT INTO Orders (OrderDate, Customer) " +
                "VALUES ({0}, {1})";
            DBContext.ExecuteCommand(insertSql, orderDate, customer);

            const string orderIdSql = "SELECT ID FROM Orders WHERE Customer = {0}";
            var result = DBContext.ExecuteQuery<int>(orderIdSql, customer).ToArray();
            int orderID = result.First();
            return orderID;
        }
        private int InsertOrder(string customer)
        {
            return InsertOrder(customer, DateTime.Now);
        }
    }
}
