package ie.rosvall.DBTesting;

import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;

import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

public class TestColumns 
{
    private static Connection con;
    
	@BeforeClass
	public static void setUpBeforeClass() 
	    throws Exception 
	{
       String connectionUrl = "jdbc:sqlserver://localhost;" +
           "databaseName=DBTesting;integratedSecurity=true;";
        try 
        {
            con = DriverManager.getConnection(connectionUrl);
        }
        catch (SQLException e) 
        {
            e.printStackTrace();
        }

	}
	
	@Before
	public void prepareTransaction()
	{
	    try
        {
            con.setAutoCommit(false);
        }
        catch (SQLException e)
        {
            Assert.fail("Cannot start transaction: " + e.getMessage());
        }
	}
	
	@After
	public void rollbackTransaction()
	{
	    try
        {
            con.rollback();
        }
        catch (SQLException e)
        {
            Assert.fail("Cannot roll back transaction: " + e.getMessage());
        }
	}
	
    //***************************************************************
    // Table Design  Match your objects?

	@Test
	public void TestOrderColumns()
	    throws SQLException
	{
        Statement stmt = con.createStatement();
        stmt.execute("SELECT ID, OrderDate, Customer FROM Orders");
	}

	@Test
    public void TestOrderLineColumns()
	    throws SQLException
    {
        Statement stmt = con.createStatement();
        stmt.execute("SELECT OrderID, ProductName, Count, Price FROM OrderLine");
    }

    @Test(expected=SQLException.class)
    public void TestInvalidQueryThrows()
        throws SQLException
    {
        Statement stmt = con.createStatement();
        stmt.execute("SELECT Nonexisting FROM Orders");

        Assert.fail("Selecting nonexisting column did not throw exception");
    }

    //***************************************************************
    // Table insert

    @Test
    public void TestInsertOrder()
        throws SQLException
    {
        Timestamp orderDate = new Timestamp(new java.util.Date().getTime());
        String customer = "Sven Rosvall";

        String insertSql = "INSERT INTO Orders (OrderDate, Customer) " +
            "VALUES (?, ?)";
        PreparedStatement stmt = con.prepareStatement(insertSql);
        stmt.setTimestamp(1, orderDate);
        stmt.setString(2, customer);
        stmt.execute();

        String querySql = "SELECT ID, OrderDate, Customer FROM Orders "+
            "WHERE OrderDate = ?";
        stmt = con.prepareStatement(querySql);
        stmt.setTimestamp(1, orderDate);
        ResultSet result = stmt.executeQuery();

        Boolean hasNext = result.next();
        Assert.assertTrue("Check first value", hasNext);

        Assert.assertEquals(customer, result.getString("Customer"));

        Timestamp orderDateFromDB = result.getTimestamp("OrderDate");
        // Fuzzy date compare. Milliseconds may differ.
        Assert.assertTrue(Math.abs(orderDate.getTime() - orderDateFromDB.getTime()) < 1000);
        
        hasNext = result.next();
        Assert.assertFalse("Check second value", hasNext);
    }

    //******************************************************************************
    // Auto-increment

    @Test
    public void TestIdentityIncrement()
        throws SQLException
    {
        // Insert two entries
        int id1 = InsertOrder("Peter Gabriel");
        int id2 = InsertOrder("Phil Collins");
        
        System.out.println("id1="+id1+", id2="+id2);

        // Verify the new entry ID is one higher than the previous.
        Assert.assertEquals(id1+1, id2);
    }

    //******************************************************************************
    // Constraints  Foreign Key

    @Test(expected=SQLException.class)
    public void TestInsertOrderLineFailsWithoutOrder() 
        throws SQLException
    {
        String sql = "INSERT INTO OrderLine (OrderID, ProductName, Count, Price) "+
            "VALUES (-1, 'Sand Grain', 1000000, 10)";
        con.prepareStatement(sql).executeUpdate();
        
        Assert.fail("Expect inserting OrderLine without order to fail.");
    }

    @Test
    public void TestInsertOrderLineSucceedsWithOrder()
        throws SQLException
    {
        int orderID = InsertOrder("Sandman");

        String sql = "INSERT INTO OrderLine (OrderID, ProductName, Count, Price) " +
            "VALUES (?, 'Sand Grain', 1000000, 10)";
        PreparedStatement stmt = con.prepareStatement(sql);
        stmt.setInt(1, orderID);
        stmt.executeUpdate();
    }

    //******************************************************************************
    // Constraints  Not Null

    @Test(expected=SQLException.class)
    public void TestInsertOrderWithoutCustomer()
        throws SQLException
    {
        Timestamp orderDate = new Timestamp(new java.util.Date().getTime());
        String insertSql = "INSERT INTO Orders (OrderDate) " +
            "VALUES (?)";
        PreparedStatement stmt = con.prepareStatement(insertSql);
        stmt.setTimestamp(1, orderDate);
        stmt.executeUpdate();
        
        Assert.fail("Expect insert without customer to fail.");
    }

    //******************************************************************************
    // Constraints  Default Values

    @Test
    public void TestInsertOrderWithoutOrderDate()
        throws SQLException
    {
        String customer = "Sid Vicious";
        Timestamp orderDate = new Timestamp(new java.util.Date().getTime());
        String insertSql = "INSERT INTO Orders (Customer) " +
            "VALUES (?)";
        PreparedStatement stmt = con.prepareStatement(insertSql);
        stmt.setString(1, customer);
        stmt.executeUpdate();

        String querySql = "SELECT ID, OrderDate, Customer FROM Orders " +
            "WHERE Customer = ?";
        stmt = con.prepareStatement(querySql);
        stmt.setString(1, customer);
        ResultSet result = stmt.executeQuery();
        
        Boolean hasNext = result.next();
        Assert.assertTrue("Check first value", hasNext);

        Assert.assertEquals(customer, result.getString("Customer"));

        Timestamp orderDateFromDB = result.getTimestamp("OrderDate");
        // Fuzzy date compare. Milliseconds may differ.
        Assert.assertTrue(Math.abs(orderDate.getTime() - orderDateFromDB.getTime()) < 1000);

        Assert.assertFalse("Check last value", result.next());
    }

    //******************************************************************************
    // Constraints  Value ranges

    @Test(expected=SQLException.class)
    public void TestInsertOrderLineWithZeroCount()
        throws SQLException
    {
        // Create an order for the order line
        String customer = "Frank Zappa";
        int orderID = InsertOrder(customer);

        String insertSql = "INSERT INTO OrderLine (OrderID, ProductName, Count, Price) " +
            "VALUES(?, 'Guitar', 0, 0)";
        PreparedStatement stmt = con.prepareStatement(insertSql);
        stmt.setInt(1, orderID);
        stmt.executeUpdate();
        
        Assert.fail("Expect insert with zero count to fail.");
    }

    //******************************************************************************
    // Stored Procedures

    @Test
    public void TestCreateOrderLineForNonExistingOrder()
        throws SQLException
    {

        String customer = "Carlos Santana";
        String product = "Plectrum";

        String searchOrderSql = "SELECT * FROM Orders WHERE Customer = ?";
        String searchOrderLineSql = "SELECT * FROM OrderLine WHERE OrderID = ?";

        // Ensure that we start without orders
        PreparedStatement stmt = con.prepareStatement(searchOrderSql);
        stmt.setString(1, customer);
        ResultSet result = stmt.executeQuery();
        Assert.assertFalse(result.next());

        // Create the order
        CallableStatement cstmt = con.prepareCall("{call CreateOrderLine(?, ?, ?, ?)}");
        cstmt.setString(1, customer);
        cstmt.setString(2, product);
        cstmt.setInt(3, 1);
        cstmt.setInt(4, 42);
        cstmt.executeUpdate();

        // Ensure that we have one order now
        stmt = con.prepareStatement(searchOrderSql);
        stmt.setString(1, customer);
        result = stmt.executeQuery();
        Assert.assertTrue(result.next());
        int orderID = result.getInt("ID");
        Assert.assertFalse(result.next());
        
        // Ensure that we have one order line
        stmt = con.prepareStatement(searchOrderLineSql);
        stmt.setInt(1, orderID);
        result = stmt.executeQuery();
        Assert.assertTrue(result.next());
        Assert.assertEquals(product, result.getString("ProductName"));
        Assert.assertEquals(42, result.getInt("Price"));      
        Assert.assertFalse(result.next());
    }
    
    //******************************************************************************
    // Constraints  Unique
    // Triggers  Cascading delete
    // Views
    // Access
    // Partitioned Tables

    //******************************************************************************
    // Helper functions

    private int InsertOrder(String customer, Timestamp orderDate) 
        throws SQLException
    {
        String insertSql = "INSERT INTO Orders (OrderDate, Customer) " +
            "VALUES (?, ?)";
        PreparedStatement stmt = con.prepareStatement(insertSql);
        stmt.setTimestamp(1, orderDate);
        stmt.setString(2, customer);
        stmt.execute();

        String querySql = "SELECT ID, OrderDate, Customer FROM Orders "+
            "WHERE OrderDate = ?";
        stmt = con.prepareStatement(querySql);
        stmt.setTimestamp(1, orderDate);
        ResultSet result = stmt.executeQuery();

        result.next();
        int orderID = result.getInt("ID");
        return orderID;
    }
    private int InsertOrder(String customer)
        throws SQLException
    {
        return InsertOrder(customer, new Timestamp(new java.util.Date().getTime()));
    }
}
